/*
 * @(#)jdemosecure.java    1.00   02/11/03
 *
 * Copyright (c) 2003 Agilis Software LLC, Fremont, CA, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Agilis Software LLC ("Confidential Information").  You shall not
 * disclose such Confidential Information and shal use it only
 * in accordance with the terms of the license agreement you entered
 * into with Agilis Software LLC.
 */

/*
 * This simple application illustrates the use of the secure license
 * key checking API of EasyLicenser for protecting the application
 * from spoofing.
 *
 * Principle:
 * Two fundamental principles are used to foil attempts at spoofing 
 * either the EasyLicenser run time library or the application code
 * that calls it:
 * 1. 3-way shared secrets: between the application, the EasyLicenser
 *    run time library, and the license key itself.  A hacker would
 *    have to successfully decode all three entities in order to be
 *    able to foil the licensing scheme.
 * 2. Obfuscated flow control.  By increasing the distance between
 *    the name of the EasyLicenser run time library API as well as
 *    the region of flow control, and the location of the license
 *    key checking API call, the application can be made arbitrarily
 *    difficult to reverse engineer.
 *
 * Mechanics:
 * Security from spoofing is provided at 4 levels:
 * 1. Using the custom key handler mechanism to implement a shared secret
 *    between the application, the EasyLicenser run time library and
 *    the license key.  The ISV embeds a custom key at the time of generating
 *    a license key.  The custom key typically contains information that is
 *    useful for license management purposes anyway - however, at its
 *    simplest level, it contains a "signature" indicating it's a key for
 *    the targeted application or application environment.  At run time,
 *    the ISV's application passes a custom key handler accompanied by
 *    a run time application context.  The application context may contain
 *    the application's signature.  The EasyLicenser invokes the custom
 *    key handler as part of its license key check, passing on the custom
 *    key embedded in the license key.  The application's implementation of
 *    the custom key handler can match the application context with the
 *    custom key to verify the identity of both, and post an event to that
 *    effect, whose result can be verified by the mainline code.
 *    If either the application code that invokes the EasyLicenser API is
 *    tampered with, or the EasyLicenser library is replaced with an impostor,
 *    the license key check fails.
 * 2. Using the EasyLicenser checkLicenseKeySecure API call to implement a
 *    dynamic handshake protocol.  The first time the application is run,
 *    it obtains a seed key token from the EasyLicenser run time library,
 *    and saves it in a secure persistent location.  Every time
 *    a license key check is to be performed, for example upon program
 *    startup, the saved key token is provided to checkLicenseKeySecure,
 *    which validates it and produces and returns an updated key token if
 *    the license key check is successful.  The application replaces the
 *    previous key token with the new token.
 *    If the application code that invokes the EasyLicenser API is
 *    tampered with, or the system clock is turned back, the EasyLicenser 
 *    run time will fail the license key check regardless of the 
 *    validity of the license key.
 * 3. Utilizing exception-based flow control with the checkLicenseKeySecure
 *    API call.  The application defines an exception handler that inherits
 *    from an EasyLicenser base exception handler class, and passes an instance
 *    of it to the EasyLicenser checkLicenseKeySecure API call.
 *    When the license key check succeeds, the API call throws the custom
 *    exception handler instead of returning control to the caller.  When the
 *    exception handler is sufficiently distant from the API call in the 
 *    application code, it becomes harder for the application to be reverse
 *    engineered.
 * 4. Invoking the EasyLicenser license key checking API call through Java
 *    reflection instead of a straight method invocation.  The application
 *    uses variable names instead of literals so as to obfuscate the API
 *    call.  This is a generic application programming technique that is 
 *    not unique to EasyLicenser.
 *
 * The sample application illustrates the use of the above techniques.
 * The successful functioning of the custom key handler technique is
 * dependent on the key you supply to the program.  Therefore, in this
 * simple example, the custom key handler doesn't enforce any event-driven
 * protocol.  In a truly paranoid application, the custom key handler would
 * post an event in the event of success.  The main program would define a
 * listener on the event, and if after the license key check, the event had
 * not been fired, that would imply a compromise in the security of one of
 * the 3 participants.
 *
 * The sample application is necessarily small and does not attempt to
 * distance the EasyLicenser API call from the success exception handler.
 * In a real application, the exception handler would be far removed from
 * the call.
 */

import java.lang.reflect.*;
import java.io.*;
import java.util.*;
import com.vs.ezlicrun.*;

public class jdemosecure {

	static final int KEY_MAXLENGTH = 1000;

	// Key token.  In a real application, this would be maintained
	// in persistent store, for example a registry entry.
	static String lastToken = null;
	static String dynClassname = "com.vs.ezlicrun.EzLicenseInfo";
	static String ezlmExceptionHandlerClassname = "com.vs.ezlicrun.EzLicExceptionBase";
	static String ezlmCustomKeyHandlerClassname = "com.vs.ezlicrun.EzLicCustomLicenseInterface";
	static String myExceptionHandlerClassname = "MyExceptionHandler";
	static String myCustomKeyHandlerClassname = "MyCustomKeyHandler";
	static String apiCallStr = "checkLicenseKeySecure";
	static String appsigStr = "jdemosecuresignature";

	public jdemosecure() {}

	/**
	 * main entry point:
	 * Accepts two command line arguments:
	 * 1. Name of a text file containing a lincense key to be validated.
	 *    The file can be any exported or published key file produced from
	 *    the EasyLicenser GUI.
	 * 2. Expected user / host name that should match what's in the key.
	 *
	 * Checks the license key and displays its license parameters if it
	 * passes the check.
	 */
	public static void main(String[] args) {
		if (args.length < 2) {
			System.out.println("*** Usage: jdemosecure key-text-file-name user-host name ***");
			System.exit(1);
		}
		String keyfile = args[0];
		String userhost = args[1];
		File kf = new File(keyfile);
		if (!kf.exists()) {
			System.out.println("*** License key file '" +
						keyfile + "' not found ***");
			System.exit(2);
		}
		String keytext = "";
		BufferedReader in = null;
		try {
			in = new BufferedReader(
					new InputStreamReader(
						new FileInputStream(kf)));
			String keyline = null;
			while ((keyline = in.readLine()) != null)
				keytext += "\n"+keyline;
	 		if ((keytext = getKey(keytext, getFilenameSuffix(keyfile))) == null)
				System.exit(1);
			if (keytext.length() > KEY_MAXLENGTH) {
				System.out.println("*** Invalid key: string too long. ***");
				System.exit(1);
			}
		} catch (Exception ex) {
			System.out.println("*** IO Error reading license key file ***");
			System.out.println("Exception: " + ex.getClass().getName() + ", Message: " + ex.getMessage());
			ex.printStackTrace();
			System.exit(2);
		} finally {
			try {
				in.close();
			} catch (Exception exx) {
			}
		}
		System.out.println("\nHello Easy Licenser World.\n");
		System.out.println("Checking license key:\n"+keytext+"\n...\n\n");
		System.out.println("Checking against user/host name: " + userhost);
		// First call illustrates the use of the key cookie and
		// custom exception handling.
		lastToken = EzLicenseInfo.createKeyCookieSeed(userhost);
		EzLicenseInfo licenseInfo = new EzLicenseInfo();
		int warnings = 0;
		try {
			licenseInfo.checkLicenseKeySecure(
						keytext,
						lastToken,
						(new MyExceptionHandler()),
						null, null,
						1000, 0, 10000,
						100, 1000,
						userhost,
						0);
		} catch (EzLicExceptionBase ex) {
			if (ex instanceof MyExceptionHandler) {
				// in practice this would be caught somewhere in the code
				// quite removed from the API call.
				lastToken = ex.getMessage();
				warnings = ex.getMessageCode();
			} else {
				System.out.println("*** License key check failed.  Error:\n\t" +
					ex.getMessage());
				System.exit(3);
			}
		}
		checkWarnings(warnings);
		displayLicenseParams(licenseInfo);
		System.out.println("\nLicense key validation successful.\n");
		System.out.println("Re-checking the key to illustrate token usage:");
		// Re-check the key, passing in the previously-returned key token,
		// to simulate a new key check at program startup.
		licenseInfo = new EzLicenseInfo();
		try {
			licenseInfo.checkLicenseKeySecure(
						keytext,
						lastToken,
						(new MyExceptionHandler()),
						null, null,
						1000, 0, 10000,
						100, 1000,
						userhost,
						0);
		} catch (EzLicExceptionBase exx) {
			if (!(exx instanceof MyExceptionHandler)) {
				System.out.println("*** Second license key check failed.  Error:\n\t" +
					exx.getMessage());
				System.exit(3);
			}
		}
		System.out.println("Re-checking the key to illustrate call w/o exception handler flow control.");
		System.out.println("This time, with a custom key handler.");
		licenseInfo = new EzLicenseInfo();
		try {
			licenseInfo.checkLicenseKeySecure(
						keytext,
						lastToken,
						null,
						(new MyCustomKeyHandler()),
						"jdemosecuresignature",
						1000, 0, 10000,
						100, 1000,
						userhost,
						0);
		} catch (EzLicExceptionBase exxx) {
			System.out.println("*** Third license key check failed.  Error:\n\t" +
				exxx.getMessage());
			System.exit(3);
		}
		System.out.println("Now do everything with dynamic invocation for further obfuscation.");
		licenseInfo = new EzLicenseInfo();
		try {
			Class dynClass = Class.forName(dynClassname);
			Class ezlmExceptionHandlerClass = Class.forName(ezlmExceptionHandlerClassname);
			Class ezlmCustomKeyHandlerClass = Class.forName(ezlmCustomKeyHandlerClassname);
			Class myExceptionHandlerClass = Class.forName(myExceptionHandlerClassname);
			Class myCustomKeyHandlerClass = Class.forName(myCustomKeyHandlerClassname);
			Object dyn = dynClass.newInstance();
			Object myexch = myExceptionHandlerClass.newInstance();
			Object mycsth = myCustomKeyHandlerClass.newInstance();
			Method methodObj = dynClass.getMethod(apiCallStr, 
										new Class[]
										{String.class,
										 String.class,
										 ezlmExceptionHandlerClass,
										 ezlmCustomKeyHandlerClass,
										 Object.class,
										 int.class, long.class, long.class,
										 int.class, long.class,
										 String.class, long.class});
			Object[] callargs = { keytext, lastToken, myexch, mycsth, 
							  appsigStr, new Integer(1000), new Long(0), 
							  new Long(10000), new Integer(100), new Long(1000),
							  userhost, new Long(0)};
			methodObj.invoke(dyn, callargs);
		} catch (Exception ex) {
			Throwable t = null;
			if (ex instanceof InvocationTargetException)
				t = ((InvocationTargetException)ex).getTargetException();
			else
				t = ex;
			if (t instanceof MyExceptionHandler) {
				System.out.println("OK.  Got success exception.");
			} else if (t instanceof Exception) {
				Exception exx = (Exception)t;
				System.out.println("*** Got unexpected exception.\n" +
							   "Exception class: " + exx.getClass().getName() +
							   "\nMessage: " + exx.getMessage());
				System.exit(1);
			} else {
				System.out.println("*** Caught unexpected throwable. ***");
				t.printStackTrace();
				System.exit(1);
			}
		}
		System.out.println("All license key validations successful.");
	}

	static void checkWarnings(int warnings)
	{
		if (warnings != 0) {
			System.out.println("\tWarnings:");
			if ((warnings & EzLicExceptionBase.EZLIC_WARN_EXPIRING) != 0)
				System.out.println("\t\tTime Limited Key Is About To Expire");
			if ((warnings & EzLicExceptionBase.EZLIC_WARN_QUOTA) != 0)
				System.out.println("\t\tQuota Limited Key Is About To Expire");
			if ((warnings & EzLicExceptionBase.EZLIC_WARN_EXPIRED) != 0)
				System.out.println("\t\tTime Limited Key Has Expired");
			if ((warnings & EzLicExceptionBase.EZLIC_WARN_QUOTA_EXCEEDED) != 0)
				System.out.println("\t\tQuota Limited Key Has Expired");
		}
	}

	static void displayLicenseParams(EzLicenseInfo licenseInfo)
	{
		System.out.println("\n\tLicense Key Information:\n\t========================\n\n");
		System.out.println("\t\tLicense key: [" + licenseInfo.getLicenseKey() + "]");
		showLicenseModel(licenseInfo);
		int bitmap = licenseInfo.getLicenseModeBitmap();
		System.out.println("\t\tLicense Mode Bitmap: " +
						bitmap);
		if ((bitmap & EzLicenseInfo.EZLIC_MODE_TIME) != 0)
			System.out.println("\t\t\tTIME LIMITED");
		if ((bitmap & EzLicenseInfo.EZLIC_MODE_OPTIONS) != 0)
			System.out.println("\t\t\tOPTIONS");
		if ((bitmap & EzLicenseInfo.EZLIC_MODE_METERED) != 0)
			System.out.println("\t\t\tMETERED");
		if ((bitmap & EzLicenseInfo.EZLIC_MODE_CUSTOM_KEY) != 0)
			System.out.println("\t\t\tCUSTOM KEY");
		if ((bitmap & EzLicenseInfo.EZLIC_MODE_CUSTOM_COOKIE) != 0)
			System.out.println("\t\t\tCUSTOM COOKIE");
		long lvalue;
		lvalue = licenseInfo.getUsageValue();
		if (lvalue > 0)
			System.out.println("\t\tUsage Limit Value: " + lvalue);
		lvalue = licenseInfo.getQuotaValue();
		if (lvalue > 0)
			System.out.println("\t\tQuota Value: " + lvalue);
		String options = licenseInfo.getOptions();
		if (options != null && !options.equals(""))
			System.out.println("\t\tOptions: [" + options + "]");
		Date expDate = licenseInfo.getExpirationDate();
		if (expDate != null)
			System.out.println("\t\tExpiration Date: " +
							expDate.toString());
		System.out.println("\t\tUser / host name: " +
						licenseInfo.getUserHostNetName());
		if (licenseInfo.getEnforce())
			System.out.println("\t\tUser / host name operating system check enforced.");
		String customCookie = licenseInfo.getCustomCookie();
		if (customCookie != null && !customCookie.equals(""))
			System.out.println("\t\tCustom Cookie: [" + customCookie + "]");
		String customKey = licenseInfo.getCustomKey();
		if (customKey != null && !customKey.equals(""))
			System.out.println("\t\tCustom Key: [" + customKey + "]");
	}

	static void showLicenseModel(EzLicenseInfo licenseInfo)
	{
		int licenseModel = licenseInfo.getLicenseModelCode();
		int licenseType = licenseInfo.getLicenseTypeCode();
		System.out.print("\t\tLicense Model:\t");
		switch(licenseModel) {
		case EzLicenseInfo.EZLIC_MODEL_USER:
			System.out.print("User\n\t\tLicense Type:\t");
			switch(licenseType) {
			case EzLicenseInfo.EZLIC_TYPE_U_USER:
				System.out.println("Named User\n");
				break;
			case EzLicenseInfo.EZLIC_TYPE_U_NODE:
				System.out.println("Node Locked\n");
				break;
			case EzLicenseInfo.EZLIC_TYPE_U_FLOAT:
				System.out.println("Floating User\n");
				break;
			default:
				System.out.println("*** Unknown license type " + licenseType + " ***");
			}
			break;
		case EzLicenseInfo.EZLIC_MODEL_SERVER:
			System.out.print("Server\n\t\tLicense Type:\t");
			switch(licenseType) {
			case EzLicenseInfo.EZLIC_TYPE_SVR_CONC:
				System.out.println("Concurrent User\n");
				break;
			case EzLicenseInfo.EZLIC_TYPE_SVR_CPU:
				System.out.println("CPU Count\n");
				break;
			case EzLicenseInfo.EZLIC_TYPE_SVR_MHZ:
				System.out.println("CPU MHZ\n");
				break;
			case EzLicenseInfo.EZLIC_TYPE_SVR_NMU:
				System.out.println("Named User\n");
				break;
			default:
				System.out.println("*** Unknown license type " + licenseType + " ***");
			}
			break;
		default:
			System.out.println("*** Unknown license model " + licenseModel + " ***");
		}
	}

	static String getFilenameSuffix(String filename) {
		int dotpos = filename.lastIndexOf('.');
		if (dotpos <= 0 || dotpos > filename.length()-4) {
			return "";
		}
		return filename.substring(dotpos+1);
	}

	static String getKey(String fileContent, String suffix) {
		int keyidx = -1;
		fileContent = fileContent.trim();
		if (suffix.equalsIgnoreCase("xml")) {
			if (!(fileContent.startsWith("<?xml") &&
			      (keyidx = fileContent.indexOf("<key id=")) > 0)) {
				System.out.println("*** Not an EasyLicenser XML file.  Aborting. ***");
				return null;
			}
			fileContent = fileContent.substring(keyidx);
			fileContent = fileContent.substring(fileContent.indexOf('>')+1);
			keyidx = fileContent.indexOf("</key>");
			if (keyidx != fileContent.lastIndexOf("</key>")) {
				System.out.println("*** File has multiple keys.  Only first key will be checked. ***");
			}
			fileContent = fileContent.substring(0, keyidx);
			return strip(fileContent);
		}
		if (suffix.equalsIgnoreCase("txt")) {
			if (fileContent.startsWith("License Key List:")) {
				if ((keyidx = fileContent.indexOf("License key:")) < 0) {
					System.out.println("*** Not an EasyLicenser-generated text file.  Aborting. ***");
					return null;
				}
				fileContent = fileContent.substring(keyidx);
				fileContent = fileContent.substring(fileContent.indexOf(":")+1);
				if (fileContent.indexOf("License key:") > 0){
					System.out.println("*** File has multiple keys.  Only first key will be checked. ***");
				}
				keyidx = fileContent.indexOf("Key Id:");
				if (keyidx > 0)
					fileContent = fileContent.substring(0, keyidx);
			}
			return strip(fileContent);
		}
		if (suffix.equalsIgnoreCase("ini")) {
			if ((fileContent = getFirstLine(fileContent)) == null)
				return null;
			if ((keyidx = fileContent.indexOf((int)'=')) <= 0) {
				System.out.println("*** Invalid property file.  Aborting. ***");
				return null;
			}
			return strip(fileContent.substring(keyidx+1));
		}
		if (suffix.equalsIgnoreCase("csv") || suffix.equalsIgnoreCase("tab")) {
			return getXsvKey(fileContent, suffix);
		}
		if (suffix.equalsIgnoreCase("key")) {
			if ((fileContent = getFirstLine(fileContent)) == null)
				return null;
			return strip(fileContent);
		}
		// else assume vanilla pure-key file -- too bad if it isn't
		return strip(fileContent);
	}

	static String getXsvKey(String fileContent, String suffix) {
		char separator = suffix.equalsIgnoreCase("csv") ? ',':'\t';
		int keyidx = -1;
		if ((fileContent = getFirstLine(fileContent)) == null)
			return null;
		if ((keyidx = fileContent.lastIndexOf((int)separator)) <= 0) {
			System.out.println("*** Invalid " + suffix +
							" file.  Aborting. ***");
			return null;
		}
		return strip(fileContent.substring(keyidx+1));
	}

	static String getFirstLine(String fileContent) {
		int lf = fileContent.indexOf("\n");
		if (lf > 0) {
			System.out.println("*** File has multiple keys.  Only first key will be checked. ***");
			return fileContent.substring(0, lf);
		} else if (lf == 0) {
			System.out.println("*** File is empty.  Nothing to do. ***");
			return null;
		}
		return fileContent;
	}

	static String strip(String instr) {
		String retstr = "";
		try {
			StringTokenizer stk = new StringTokenizer(instr," \n\t");
			while (stk.hasMoreTokens()) {
				String tokstr = stk.nextToken();
				retstr += tokstr;
			}
		} catch (Exception ex) {
			System.out.println("*** Got exception: " + ex.getMessage());
		}
		return retstr;
	}
	
}
// Custom exception handler for signalling success
class MyExceptionHandler extends EzLicExceptionBase {
	MyExceptionHandler() {}
}
// Custom key handler
class MyCustomKeyHandler implements EzLicCustomLicenseInterface {
	
	public int checkCustomKeyCode(String valStr, Object ctx) 
			throws EzLicExceptionBase
	{
		// checks for whether the run time context matches the custom
		// key.  Since a variety of keys can be fed into the program,
		// we simply print the status of the match here instead of
		// throwing an exception.
		String appname = (String)ctx;
		if (!appname.equals(valStr)) {
			System.out.println("*** Signature mismatch.  "+appname+" vs. " + valStr + "***");
			System.out.println("Possible compromise in 3-way secret.");
			// should throw an exception here
		} else
			System.out.println("Signatures match and all is well.");
			// should also post a success event here
		return 0;
	}

}
